<?php

/**
 * This file contains QUI\Auth\WalletConnect\Auth
 */

namespace QUI\Auth\WalletConnect;

use Elliptic\EC;
use kornrunner\Keccak;

use QUI;
use QUI\Users\AbstractAuthenticator;
use QUI\Users\User;

/**
 * Class Auth
 *
 * Authentication handler for WalletConnect authentication
 *
 * @package QUI\Auth\WalletConnect
 */
class Auth extends AbstractAuthenticator
{
    /**
     * User that is to be authenticated
     *
     * @var User
     */
    protected $User = null;

    /**
     * Auth Constructor.
     *
     * @param string|array|integer $user - name of the user, or user id
     */
    public function __construct($user = '')
    {
        if (!empty($user)) {
            try {
                $this->User = QUI::getUsers()->getUserByName($user);
            } catch (\Exception $Exception) {
                $this->User = QUI::getUsers()->getNobody();
            }
        }
    }

    /**
     * @param null|QUI\Locale $Locale
     * @return string
     */
    public function getTitle($Locale = null): string
    {
        if (\is_null($Locale)) {
            $Locale = QUI::getLocale();
        }

        return $Locale->get('quiqqer/authwalletconnect', 'authwalletconnect.title');
    }

    /**
     * @param null|QUI\Locale $Locale
     * @return string
     */
    public function getDescription($Locale = null): string
    {
        if (\is_null($Locale)) {
            $Locale = QUI::getLocale();
        }

        return $Locale->get('quiqqer/authwalletconnect', 'authwalletconnect.description');
    }

    /**
     * Authenticate the user
     *
     * @param array|string $authData
     *
     * @throws QUI\Auth\WalletConnect\Exception
     */
    public function auth($authData)
    {
        $message   = $authData['message'];
        $signature = $authData['signature'];
        $address   = $authData['address'];

        try {
            $verify = QUI\Auth\WalletConnect\Auth::verifySignature($message, $signature, $address);
        } catch (\Exception $Exception) {
            throw new QUI\Auth\WalletConnect\Exception('Authentication failed');
        }

        if (!$verify) {
            throw new QUI\Auth\WalletConnect\Exception('Authentication failed');
        }

        // get user

        // if no user, create it

        // get user

        // set qui session

        // change nonce - security reasons
    }

    /**
     * @param $address
     * @throws QUI\Auth\WalletConnect\Exception
     * @throws QUI\Users\Exception
     * @throws QUI\Exception
     */
    public function setSession($address)
    {
        try {
            $userId = QUI\WalletConnect\WalletConnect::getUserIdByEthAddress($address);
        } catch (\Exception $Exception) {
            throw new QUI\Auth\WalletConnect\Exception('Authentication failed');
        }

        $User = null;

        if (!empty($userId)) {
            try {
                $User = QUI::getUsers()->get($userId);
            } catch (QUI\Exception $Exception) {
            }
        }

        // user has no started registration
        if ($User === null) {
            $FrontendUsers = new QUI\FrontendUsers\Handler();
            $Registrars    = $FrontendUsers->getRegistrars();

            $Registrars = $Registrars->filter(function ($R) {
                return $R instanceof QUI\Registration\WalletConnect\Registrar;
            });

            /* @var $Registrar QUI\Registration\WalletConnect\Registrar */
            $Registrar = $Registrars->get(0);

            $_POST['registrar']    = $Registrar->getHash();
            $_POST['registration'] = 1;

            $Registration = new QUI\FrontendUsers\Controls\Registration();
            $Registration->setAttribute('Registrar', $Registrar);
            $Registration->register();

            // check if name already exists
            $User = $Registration->getRegisteredUser();

            if (!QUI::getUsers()->usernameExists($address)) {
                $User->setAttribute('username', $address);
                $User->save(QUI::getUsers()->getSystemUser());
            }

            QUI::getDataBase()->update(
                QUI\WalletConnect\WalletConnect::table(),
                [
                    'userId' => $User->getId(),
                    'nonce'  => \uniqid() // reset nonce, because of security
                ],
                ['address' => $address]
            );
        }

        $Session = QUI::getSession();
        $Session->set('uid', $User->getId());
        $Session->set('username', $User->getName());
        $Session->set('auth-globals', 1);

        QUI::getUsers()->login();
    }

    /**
     * Return the user object
     *
     * @return QUI\Interfaces\Users\User
     */
    public function getUser()
    {
        return $this->User;
    }

    /**
     * Return the quiqqer user id
     *
     * @return integer|boolean
     */
    public function getUserId()
    {
        return $this->User->getId();
    }

    /**
     * @return QUI\Control
     */
    public static function getLoginControl()
    {
        return new QUI\Auth\WalletConnect\Controls\Login();
    }

    /**
     * @return QUI\Control
     */
    public static function getPasswordResetControl(): ?QUI\Control
    {
        return null;
    }

    /**
     * @param $message
     * @param $signature
     * @param $address
     *
     * @return false
     *
     * @throws \Exception
     */
    public static function verifySignature($message, $signature, $address): bool
    {
        $msgLen = \strlen($message);

        $hash = Keccak::hash(
            "\x19Ethereum Signed Message:\n{$msgLen}{$message}",
            256
        );

        $sign = [
            "r" => \substr($signature, 2, 64),
            "s" => \substr($signature, 66, 64)
        ];

        $recId = \ord(\hex2bin(\substr($signature, 130, 2))) - 27;

        if ($recId != ($recId & 1)) {
            return false;
        }

        $EC     = new EC('secp256k1');
        $pubKey = $EC->recoverPubKey($hash, $sign, $recId);
        $pub    = self::pubKeyToAddress($pubKey);

        // damn workaround
        return strtolower($address) == strtolower($pub);
        //return $address == $pub;
    }

    /**
     * @param $pubKey
     * @return string
     *
     * @throws \Exception
     */
    public static function pubKeyToAddress($pubKey): string
    {
        return "0x".substr(Keccak::hash(substr(hex2bin($pubKey->encode("hex")), 1), 256), 24);
    }
}
